!= ե륪ץֹ
!
!= Unit number handling at file open
!
! Authors::   Yasuhiro MORIKAWA
! Version::   $Id: dc_iounit.f90,v 1.2 2007/07/11 13:58:41 morikawa Exp $
! Tag Name::  $Name: gt4f90io-20071009 $
! Copyright:: Copyright (C) GFD Dennou Club, 2007. All rights reserved.
! License::   See COPYRIGHT[link:../../COPYRIGHT]
!

module dc_iounit
  !
  != ե륪ץֹ
  !
  != Unit number handling at file open
  !
  ! <b>Note that Japanese and English are described in parallel.</b>
  !
  ! ե򥪡ץ󤹤ݤ˻ѤֹνԤޤ. 
  ! ֥롼 FileOpen ˥ե̾ȥץΥ⡼ɤͿ뤳Ȥ, 
  ! Ѳǽֹõ, 
  ! ǥե⡼ɤ˹碌ƥץ󤷤Ƥ, ֹ֤ޤ. 
  ! Υ⥸塼, ֹδեɤ߼/񤭹߲ǽ
  ! ʤɤγǧȤΤΥɤǲޤ.
  !
  ! Unit number is handled when a file is opened.
  ! Subroutine "FileOpen" receives filename and open mode firstly, and
  ! searches an available unit number internally, and
  ! opens the file according to the open mode internally, and
  ! returns the unit number finally.
  ! This module simplifies source codes for unit number management or
  ! readable/writable check.
  !
  !== Procedures List
  !
  ! FileOpen           :: ե̾ȥ⡼ɤͿ뤳Ȥǥե
  !                       ץ, ֹ֤ޤ.
  ! ------------       :: ------------
  ! FileOpen           :: A file is opened and unit number is returned with
  !                       a filename and mode
  !
  !== Usage
  !
  ! Υ⥸塼󶡤륵֥롼 FileOpen 
  ! ե̾ȥץΥ⡼ɤͿƤ.
  ! ȥե뤬ץ󤵤,  *unit* ֹ椬֤ޤ.
  ! ֹѤ, եƤɤ߹ߤ
  ! եؤν񤭹ߤԤäƤ. ɤ߹ߤ񤭹ߤ䥯ˤ 
  ! Fortran Ȥ߹ߤ READ ʸ WRITE ʸ, CLOSE ʸѤƤ.
  !
  ! Give filename and open mode to subroutine "FileOpen" provided by
  ! this module.  Then the file is opened and unit number is returned
  ! to an argument *unit*.  Using the unit number, read the contents
  ! of the file or write in the file.  Use Fortran built-in READ,
  ! WRITE, and CLOSE statements for read/write and close.
  !
  !=== Example
  !
  !    program dc_iounit_sample
  !      use dc_types, only: TOKEN
  !      use dc_iounit, only: FileOpen
  !      implicit none
  !      integer:: unit00, unit01
  !      character(TOKEN):: char
  !      character(TOKEN):: filename = 'dc_iounit_sample.nml'
  !      integer:: int
  !      namelist /dc_iounit_sample_nml/ char, int
  !    continue
  !    
  !      call FileOpen(unit00, file = filename, mode = 'w')
  !      write(unit00, *) '&dc_iounit_sample_nml'
  !      write(unit00, *) '  char = "hogehoge",'
  !      write(unit00, *) '  int = 123'
  !      write(unit00, *) '/'
  !      close(unit00)
  !    
  !      call FileOpen(unit01, file = filename, mode = 'r')
  !      read(unit01, nml = dc_iounit_sample_nml)
  !      close(unit01)
  !      write(0, nml = dc_iounit_sample_nml)
  !    
  !    end program dc_iounit_sample
  !
  implicit none
  private

  public:: FileOpen

  character(*), parameter:: version = &
    & '$Name: gt4f90io-20071009 $' // &
    & '$Id: dc_iounit.f90,v 1.2 2007/07/11 13:58:41 morikawa Exp $'

  interface FileOpen
    module procedure FileOpen
  end interface

contains

  subroutine FileOpen( &
    & unit, file, mode, &
    & err )
    !
    ! ե̾ *file* , ץ󤹤ݤΥ⡼ɤ *mode* ؤ
    ! Ϳ뤳Ȥ, ե򥪡ץ, ֹ *unit* ֤ޤ. 
    ! *mode* ˤϰʲʸꤷޤ. ά "r" ꤵ줿
    ! Ȥߤʤޤ. 
    !
    ! "r"  :: եɤ߹ߥ⡼ɤǥץ󤷤ޤ.
    ! "w"  :: ե񤭹ߥ⡼ɤǥץ󤷤ޤ. 
    !         ץ˥ե뤬Ǥ¸ߤƤФƤˤޤ.
    ! "a"  :: ե񤭹ߥ⡼ɤǥץ󤷤ޤ. 
    !         Ϥϥեɲäޤ.
    ! "rw" :: եɤ߽ξѥ⡼ɤǥץ󤷤ޤ. 
    !         ץ˥ե뤬Ǥ
    !         ¸ߤƤФƤˤޤ.
    ! "ra" :: եɤ߽ξѥ⡼ɤǥץ󤷤ޤ. 
    !         ץ˥ե뤬Ǥ
    !         ¸ߤƤɤ߽񤭰֤ե˥åȤޤ.
    !
    ! ե뤬 *mode* ǻꤵ⡼ɤǳʤ, ץ
    ! λޤ.  *err* Ϳ, ץ϶λ, 
    !  *err*  .true. , *unit*  -1 ޤ. 
    !
    ! Filename is given to *file*, and open mode is given to *mode*, 
    ! then the file is opened and unit number is returned.
    !
    ! "r"  :: A file is opened with read-only mode
    ! "w"  :: A file is opened with writable mode. 
    !         If a file is exist already, the contest of the file is emptied.
    ! "a"  :: A file is opened with writable mode.
    !         Output is appended at the end of the file.
    ! "rw" :: A file is opened with read/write mode. 
    !         If a file is exist already, the contest of the file is emptied. 
    ! "ra" :: A file is opened with read/write mode. 
    !         If a file is exist already, 
    !         a position of read/write is set at the end of the file.
    !
    ! If the file can not be opened with the mode, the program aborts. 
    ! If this *err* argument is given, .true. is substituted to *err* and
    ! -1 is substituted to *unit* and the program does not abort. 
    !
    use dc_types, only: STRING, TOKEN
    use dc_trace, only: BeginSub, EndSub
    use dc_error, only: StoreError, DC_NOERR, &
      & DC_EFILENAMEEMPTY, DC_EBADFILEOPMODE, DC_ENOUNITNUM, &
      & DC_ENOFILEEXIST, DC_ENOFILEREAD, DC_ENOFILEWRITE
    use dc_present, only: present_and_not_empty
    use dc_string, only: toChar, toLower
    implicit none
    integer, intent(out):: unit
    character(*), intent(in):: file
    character(*), intent(in), optional:: mode
    logical, intent(out), optional:: err

    !-----------------------------------
    !  ѿ
    !  Work variables
    integer, parameter:: max_unit = 99
                              ! NAMELIST ե򥪡ץ󤹤뤿
                              ! ֹκ. Fortran ǻѲǽ
                              ! ϰ (099) Τ, 
                              ! ͤꤵƤ.
                              ! 
                              ! Maximum unit number for open of
                              ! NAMELIST file. An maximum
                              ! value within the bounds of available number
                              ! in Fortran (0 - 99) is specified.
    integer, parameter:: min_unit = 0
                              ! NAMELIST ե򥪡ץ󤹤뤿
                              ! ֹκǾ. Fortran ǻѲǽ
                              ! ϰ (099) Τ, 
                              ! ǾͤꤵƤ.
                              ! 
                              ! Minimum unit number for open of
                              ! NAMELIST file. An minimum
                              ! value within the bounds of available number
                              ! in Fortran (0 - 99) is specified.
    character(TOKEN):: open_mode
    integer:: unit_work
    logical:: unit_exist_flag, unit_opend_flag
    logical:: file_exist_flag
    integer:: iostat
    integer:: stat
    character(STRING):: cause_c
    character(*), parameter:: subname = 'FileOpen'
  continue
    call BeginSub(subname, version)
    stat = DC_NOERR
    cause_c = ''
    unit = -1

    !-----------------------------------------------------------------
    !  ץʥΥå
    !  Check optional arguments
    !-----------------------------------------------------------------
    if (present_and_not_empty(mode)) then
      open_mode = mode
    else
      open_mode = 'r'
    end if
    call toLower(open_mode)

    !-----------------------------------------------------------------
    !  Υå
    !  Validation of arguments
    !-----------------------------------------------------------------
    if ( trim(file) == '' ) then
      stat = DC_EFILENAMEEMPTY
      goto 999
    end if

    !----------------------------------------------------------------
    !  Ѳǽֹõ
    !  Search available unit number
    !----------------------------------------------------------------
    unit_work = max_unit
    do
      inquire(unit=unit_work, exist=unit_exist_flag, opened=unit_opend_flag)
      if (unit_exist_flag .and. .not. unit_opend_flag) then
        exit
      endif
      unit_work = unit_work - 1
      if (unit_work < min_unit) then
        cause_c = toChar(min_unit) // ' - ' // toChar(max_unit)
        stat = DC_ENOUNITNUM
        goto 999
      end if
    enddo

    !----------------------------------------------------------------
    !  ⡼ɤν񼰤Υå
    !  Check form of mode
    !----------------------------------------------------------------
    select case( trim(open_mode) )
    case ('r', 'w', 'rw', 'a', 'ra')
    case default
      cause_c = open_mode
      stat = DC_EBADFILEOPMODE
      goto 999
    end select

    !----------------------------------------------------------------
    !  ե¸ߤΥå
    !  Check existance of a file
    !----------------------------------------------------------------
    select case( trim(open_mode) )
    case ('r')
      inquire(file=file, exist=file_exist_flag)
      if (.not. file_exist_flag) then
        cause_c = file
        stat = DC_ENOFILEEXIST
        goto 999
      end if
    end select

    !----------------------------------------------------------------
    !  եɤ߹߲ǽΥå
    !  Check readable of a file
    !----------------------------------------------------------------
    select case( trim(open_mode) )
    case ('r')
      open(unit=unit_work, iostat=iostat, &
        & file=file, status='OLD', action='READ')
      if (.not. iostat == 0) then
        cause_c = file
        stat = DC_ENOFILEREAD
        goto 999
      end if
      close(unit=unit_work)
    end select

    !----------------------------------------------------------------
    !  եν񤭹߲ǽΥå
    !  Check writable of a file
    !----------------------------------------------------------------
    select case( trim(open_mode) )
    case ('w', 'a', 'rw', 'ra')
      open(unit=unit_work, iostat=iostat, &
        & file=file, status='UNKNOWN', action='WRITE')
      if (.not. iostat == 0) then
        cause_c = file
        stat = DC_ENOFILEWRITE
        goto 999
      end if
      close(unit=unit_work)
    end select

    !----------------------------------------------------------------
    !  ե륪ץ
    !  Open a file
    !----------------------------------------------------------------
    select case( trim(open_mode) )
    case ('r')
      open(unit=unit_work, file=file, &
        & status='OLD', action='READ')

    case ('w')
      open(unit=unit_work, file=file, &
        & status='REPLACE', action='WRITE')

    case ('rw')
      open(unit=unit_work, file=file, &
        & status='REPLACE', action='READWRITE')

    case ('a')
      open(unit=unit_work, file=file, &
        & status='UNKNOWN', position='APPEND', action='WRITE')

    case ('ra')
      open(unit=unit_work, file=file, &
        & status='UNKNOWN', position='APPEND', action='READWRITE')

    end select

    unit = unit_work

999 continue
    call StoreError(stat, subname, err, cause_c)
    call EndSub(subname)
  end subroutine FileOpen

end module dc_iounit
